<div layout-fill>
<md-toolbar md-scroll-shrink class="dashboard-subnav">
   <div class="md-toolbar-tools">
      <div layout="row" flex class="fill-height">
         <div class="md-toolbar-item md-breadcrumb"></div>
         <span ng-if="menu.currentPage.name !== menu.currentSection.name">
         <span hide-sm hide-md><a href="#{{menu.currentSection.url}}">{{menu.currentSection.name}}</a></span>
         <span class="menu-separator-icon" style="padding: 0 10px;" hide-sm hide-md>
         <img style="height: 12px;" src="assets/img/docArrow.png" alt="" aria-hidden="true">
         </span>
         </span>
         <span style="display: inline-block;">{{(menu.currentPage | humanizeDoc) || 'Kubernetes' }}</span>
         <span flex></span>
         <div class="md-toolbar-item md-tools" layout="row">
            <div layout="column" class="selectSubPages">
               <md-select ng-model="page" placeholder="Views" class="selectTitle">
                  <md-optgroup label="Dashboard">
                  <md-option id="{{subpage.id}}" ng-value="subpage.value" ng-repeat="subpage in subpages | filter: {category: 'dashboard' }">{{subpage.name}}</md-option>
                  </md-option-group>
               </md-select>
            </div>
         </div>
      </div>
   </div>
</md-toolbar>
</div>
