<div>
  <md-content>
    <div class="server-overview" layout="column">
      <!-- subheader -->
      <div class="group-heading" layout="row">
        <div class="label">{{selectedGroupByName | ucfirst}}: <span class="bold">{{ (groupName) || "blank"  }}</span></div>
      </div>
      <!-- render group data -->
      <div ng-include="'components/dashboard/views/partials/groupItem.html'"></div>
      <div class="footer">
        <!-- Alternate box inside a box -->
        <div ng-if="group.kind == 'grouping'">
          {{group.kind}}
          <div ng-repeat="(groupName,group) in group.items" ng-include="'components/dashboard/views/partials/groupBox.html'"></div>
        </div>
        <md-divider></md-divider>
      </div>
    </div>
  </md-content>
</div>
