apiVersion: v1
kind: Config
current-context: controller-manager-to-{{ cluster_name }}
preferences: {}
clusters:
- cluster:
    certificate-authority: {{ kube_cert_dir }}/ca.crt
    server: https://{{ groups['masters'][0] }}:443
  name: {{ cluster_name }}
contexts:
- context:
    cluster: {{ cluster_name }}
    user: controller-manager
  name: controller-manager-to-{{ cluster_name }}
users:
- name: controller-manager
  user:
    token: {{ controller_manager_token }}
