/**
 * Angular directive to convert JSON into human readable table. Inspired by https://github.com/marianoguerra/json.human.js.
 * @version v1.2.1 - 2014-12-22
 * @link https://github.com/yaru22/angular-json-human
 * @author Brian Park <yaru22@gmail.com>
 * @license MIT License, http://www.opensource.org/licenses/MIT
 */
/**
 * DISCLAIMER: This CSS is copied from https://github.com/marianoguerra/json.human.js
 */

.jh-root,
.jh-type-object,
.jh-type-array,
.jh-key,
.jh-value,
.jh-root tr {
  -webkit-box-sizing: border-box; /* Safari/Chrome, other WebKit */
     -moz-box-sizing: border-box; /* Firefox, other Gecko */
          box-sizing: border-box; /* Opera/IE 8+ */
}

.jh-key,
.jh-value {
  margin: 0;
  padding: 0.2em;
}

.jh-value {
  border-left: 1px solid #ddd;
}

.jh-type-bool,
.jh-type-number {
  font-weight: bold;
  text-align: center;
  color: #5286BC;
}

.jh-type-string {
  font-style: italic;
  color: #839B00;
}

.jh-array-key {
  font-style: italic;
  font-size: small;
  text-align: center;
}

.jh-object-key,
.jh-array-key {
  color: #444;
  vertical-align: top;
}

.jh-type-object > tbody > tr:nth-child(odd),
.jh-type-array > tbody > tr:nth-child(odd) {
  background-color: #f5f5f5;
}

.jh-type-object > tbody > tr:nth-child(even),
.jh-type-array > tbody > tr:nth-child(even) {
  background-color: #fff;
}

.jh-type-object,
.jh-type-array {
  width: 100%;
  border-collapse: collapse;
}

.jh-root {
  border: 1px solid #ccc;
  margin: 0.2em;
}

th.jh-key {
  text-align: left;
}

.jh-type-object > tbody > tr,
.jh-type-array > tbody > tr {
  border: 1px solid #ddd;
  border-bottom: none;
}

.jh-type-object > tbody > tr:last-child,
.jh-type-array > tbody > tr:last-child {
  border-bottom: 1px solid #ddd;
}

.jh-type-object > tbody > tr:hover,
.jh-type-array > tbody > tr:hover {
  border: 1px solid #F99927;
}

.jh-empty {
  font-style: italic;
  color: #999;
  font-size: small;
}
