<ul class="kubernetes-ui-menu">
  <li ng-repeat="section in menu.sections" class="parent-list-item" ng-class="{'parentActive' : isSectionSelected(section)}">
    <h2 class="menu-heading" ng-if="section.type === 'heading'" id="heading_{{ section.name | nospace }}">
      {{section.name}}
    </h2>
    <menu-link section="section" ng-if="section.type === 'link'"></menu-link>
    <menu-toggle section="section" ng-if="section.type === 'toggle'"></menu-toggle>
    <ul ng-if="section.children" class="menu-nested-list">
      <li ng-repeat="child in section.children" ng-class="{'childActive' : isSectionSelected(child)}">
        <menu-toggle section="child"></menu-toggle>
      </li>
    </ul>
  </li>
</ul>
