<div dashboard-header></div>
<div class="dashboard">
  <div ng-controller="ListServicesCtrl" style="padding:25px;" class="list-pods">
  <md-table headers="headers" content="content" sortable="sortable" filters="search" custom-class="custom" thumbs="thumbs" count="count" on-select="go(data)"></md-table>
  </div>
</div>
<div dashboard-footer></div>
