<div dashboard-header></div>
<div class="dashboard">
  <div ng-controller="PodCtrl" layout="column" class="body-wrapper pod">
    <div class="detail">

      <div class="back">
        <md-button class="md-default-theme" ng-click="doTheBack()">‹ BACK</md-button>
      </div>

      <div class="heading">
        <span class="label">Name:</span>
        <span>{{pod.metadata.name}}</span>
      </div>


      <table class="align-top">
        <tbody>
          <tr>
            <td class="name">Status</td>
            <td class="value">
              {{pod.status.phase}} on <a ng-href="#/dashboard/groups/host/selector/host={{pod.spec.nodeName}}">{{pod.spec.nodeName}}</a>
            </td>
          </tr>

          <tr>
            <td class="name">Created</td>
            <td class="value">
              {{pod.metadata.creationTimestamp | date:'medium'}}
            </td>
          </tr>

          <tr>
            <td class="name">Host Networking</td>
            <td class="value">
              {{pod.spec.nodeName}}/{{pod.status.hostIP}}
            </td>
          </tr>

          <tr>
            <td class="name">Pod Networking</td>
            <td class="value">
              {{pod.status.podIP}}
              <span ng-repeat="container in pod.spec.containers">
                <span ng-repeat="port in container.ports">
                  <span ng-show="port.containerPort">
                    {{port.name}}: {{port.containerPort}}{{$last ? '' : ', '}}
                  </span>
               </span>
              </span>
            </td>
          </tr>

          <tr>
            <td class="name">Labels</td>
            <td class="value">
              <div ng-repeat="(label, value) in pod.metadata.labels">
                {{label}}: {{value}}
              </div>
            </td>
          </tr>

          <tr>
            <td class="name">Containers</td>
            <td class="value">

              <table class="containerTable">
                <tr>
                  <td>Name</td>
                  <td>Image</td>
                  <td>Ready</td>
                  <td>Restarts</td>
                  <td>State</td>
                </tr>
                <tr ng-repeat="container in pod.status.containerStatuses">
                  <td>{{container.name}}</td>
                  <td>{{container.image}}</td>
                  <td>{{container.ready}}</td>
                  <td>{{container.restartCount}}</td>
                  <td>
                    <div ng-show="container.state.running">
                      <div class="run-state">Running</div>
                      <div class="run-messages">Started: {{container.state.running.startedAt | date:'medium'}}</div>
                    </div>
                    <div ng-show="container.state.waiting">
                      <div class="run-state">Waiting</div>
                      <div class="run-messages" ng-show="container.state.waiting.reason">Reason: {{container.state.waiting.reason}}</div>
                    </div>
                    <div ng-show="container.state.termination">
                      <div class="run-state">Terminated</div>
                      <div class="run-messages" ng-show="container.state.termination.reason">Reason: {{container.state.termination.reason}}</div>
                      <div class="run-messages" ng-show="container.state.termination.message">Message: {{container.state.termination.message}}</div>
                      <div class="run-messages">Exit Code: {{container.state.termination.exitCode}}</div>
                      <div class="run-messages" ng-show="container.state.termination.signal">Signal: {{container.state.termination.signal}}</div>
                      <div class="run-messages">Started: {{container.state.termination.startedAt | date:'medium'}}</div>
                      <div class="run-messages">Finished: {{container.state.termination.finishedAt | date:'medium'}}</div>
                    </div>
                  </td>
                </tr>
              </table>


            </td>
          </tr>


        </tbody>
      </table>



    </div>
  </div>
</div>
<div dashboard-footer></div>
