// Package accounts contains functionality for working with Object Storage
// account resources. An account is the top-level resource the object storage
// hierarchy: containers belong to accounts, objects belong to containers.
//
// Another way of thinking of an account is like a namespace for all your
// resources. It is synonymous with a project or tenant in other OpenStack
// services.
package accounts
