// Package rackconnect allows Rackspace cloud accounts to leverage version 3 of
// RackConnect, Rackspace's hybrid connectivity solution connecting dedicated
// and cloud servers.
package rackconnect
