/*!
 * Angular Material Design
 * https://github.com/angular/material
 * @license MIT
 * v0.8.1
 */
.md-select-menu-container {
  position: absolute;
  left: 0;
  top: 0;
  z-index: 99;
  opacity: 0; }
  .md-select-menu-container:not(.md-clickable) {
    pointer-events: none; }
  .md-select-menu-container md-progress-circular {
    display: table;
    margin: 24px auto !important; }
  .md-select-menu-container.md-active {
    opacity: 1; }
    .md-select-menu-container.md-active md-select-menu {
      transition: -webkit-transform all 0.4s cubic-bezier(0.25, 0.8, 0.25, 1);
      transition: transform all 0.4s cubic-bezier(0.25, 0.8, 0.25, 1);
      transition-duration: 200ms; }
      .md-select-menu-container.md-active md-select-menu > * {
        opacity: 1;
        transition: all 0.3s cubic-bezier(0.55, 0, 0.55, 0.2);
        transition-duration: 200ms;
        transition-delay: 100ms; }
  .md-select-menu-container.md-leave {
    opacity: 0;
    transition: all 0.3s cubic-bezier(0.55, 0, 0.55, 0.2);
    transition-duration: 250ms; }

md-select {
  display: inline-block;
  margin-top: 20px; }
  md-select[disabled]:hover {
    cursor: default; }
  md-select:not([disabled]):hover {
    cursor: pointer; }
  md-select:not([disabled]):focus .md-select-label {
    border-bottom: 2px solid;
    padding-bottom: 7px; }

.md-select-label {
  display: -webkit-flex;
  display: -ms-flexbox;
  display: flex;
  -webkit-align-items: center;
      -ms-flex-align: center;
          align-items: center;
  padding-top: 9px;
  padding-right: 32px;
  padding-bottom: 8px;
  border-bottom: 1px solid;
  font-size: 0.875em;
  line-height: 0.8em;
  position: relative;
  box-sizing: border-box;
  min-width: 64px; }
  .md-select-label .md-select-icon:after {
    content: '\25BC';
    display: -webkit-flex;
    display: -ms-flexbox;
    display: flex;
    -webkit-align-items: center;
        -ms-flex-align: center;
            align-items: center;
    position: absolute;
    height: 100%;
    top: 0px;
    right: 2px;
    speak: none;
    -webkit-transform: scaleY(0.6) scaleX(1);
            transform: scaleY(0.6) scaleX(1); }

md-select-menu {
  display: -webkit-flex;
  display: -ms-flexbox;
  display: flex;
  -webkit-flex-direction: column;
      -ms-flex-direction: column;
          flex-direction: column;
  box-shadow: 0px 2px 5px 0 rgba(0, 0, 0, 0.26);
  -webkit-transform-origin: 0 0;
          transform-origin: 0 0;
  -webkit-transform: scale(1);
          transform: scale(1);
  max-height: 256px;
  overflow-y: scroll; }
  md-select-menu.md-reverse {
    -webkit-flex-direction: column-reverse;
        -ms-flex-direction: column-reverse;
            flex-direction: column-reverse; }
  md-select-menu:not(.md-overflow) md-content {
    padding-top: 8px;
    padding-bottom: 8px; }
  md-select-menu md-content {
    min-width: 136px; }
  md-select-menu > * {
    opacity: 0; }

md-option {
  cursor: pointer;
  position: relative;
  display: -webkit-flex;
  display: -ms-flexbox;
  display: flex;
  -webkit-align-items: center;
      -ms-flex-align: center;
          align-items: center;
  width: auto;
  padding: 0 16px 0 16px;
  height: 48px; }
  md-option .md-text {
    width: auto;
    white-space: nowrap;
    overflow: hidden;
    text-overflow: ellipsis;
    font-size: 0.875em; }

md-optgroup {
  display: block; }
  md-optgroup label {
    display: block;
    font-size: 0.75em;
    text-transform: uppercase;
    padding: 16px 8px; }
  md-optgroup md-option {
    padding-left: 24px; }
