// Package containers contains functionality for working with Object Storage
// container resources. A container serves as a logical namespace for objects
// that are placed inside it - an object with the same name in two different
// containers represents two different objects.
//
// In addition to containing objects, you can also use the container to control
// access to objects by using an access control list (ACL).
package containers
